coverage run manage.py test

echo "To see test coverage:"
echo ""
echo 'coverage report --omit="annalist/tests/*","*/site-packages/*"'
echo 'coverage report --include=<file> --show-missing'
echo "coverage html ..."
echo ""
echo ""
