from __future__ import unicode_literals
from __future__ import absolute_import, division, print_function

"""
annalist_manager tests
"""

__author__      = "Graham Klyne (GK@ACM.ORG)"
__copyright__   = "Copyright 2018, G. Klyne"
__license__     = "MIT (http://opensource.org/licenses/MIT)"

def get_source_root():
	import os
	here     = os.path.dirname(__file__)
	src_root = os.path.normpath(os.path.join(here, "../../"))
	return src_root

# End.
